
/* A5SES is released under the zlib license.
 * See LICENSE.txt for more info.
 */

/// @file Display.hpp
/// @ingroup Display
/// Contains display related classes

#ifndef A5SES_DISPLAY_HPP_INCLUDED
#define A5SES_DISPLAY_HPP_INCLUDED

#include "common.hpp"
#include "Bitmap.hpp"
#include "Event.hpp"

// For Windows vsync disable hack
#include <allegro5/allegro_opengl.h>
#include <cstdio>

namespace a5
{

/// @ingroup Display
/// Object representing a display (screen or window)
class Display : public Bitmap, public Event_Source
{
	protected:
		/// Pointer to the Allegro display
		ALLEGRO_DISPLAY *display;

	public:
		/// Creates the display as a window
		static const int Windowed = ALLEGRO_WINDOWED;

		/// Creates the display fullscreen
		static const int Fullscreen = ALLEGRO_FULLSCREEN;

		/// Allows the display to be resized
		/// Only effective in Windowed mode without NoFrame
		static const int Resizable = ALLEGRO_RESIZABLE;

		/// Removes the window border
		/// Only effective in Windowed mode
		static const int NoFrame = ALLEGRO_NOFRAME;

		/// Generates ExposeEvent events when parts of the screen are "exposed"
		static const int GenerateExposeEvents = ALLEGRO_GENERATE_EXPOSE_EVENTS;

		/// Makes Flip() wait for VSYNC before drawing
		static const int VSync = 0x80000000;

		/// @ingroup Event Display
		/// Display event object
		class Event : public a5::Event
		{
			public:
				/// Event subtypes
				enum Sub
				{
					Close = 0x00000001,
					Resize = 0x00000002,
					Expose = 0x00000004,
					Lost = 0x00000008,
					Found = 0x00000010,
					SwitchOut = 0x00000020,
					SwitchIn = 0x00000040,
					Any = 0xFFFFFFFF,
				};

				static const int TypeID = 0x00000008;

				/// Creates a Display event
				Event(const Event_Source *source_, int subtype_)
				{
					this->source = source_;
					this->subtype = subtype_;
				}

				virtual int Type()
				{
					return TypeID;
				}
		};

		/// Creates a display with the specified width, height and flags
		/// flags is compatable with allegro5
		Display(point width, point height, int flags)
		{
			_init();

			bool vsync = flags & VSync;
			flags &= ~VSync;

			al_set_new_display_option(ALLEGRO_VSYNC, vsync ? 1 : 2, ALLEGRO_SUGGEST);
			al_set_new_display_flags(flags);

			this->display = al_create_display(width, height);
			this->bitmap = al_get_backbuffer();

			flags = al_get_display_flags();

// Remove this when vsync control is supported for opengl in windows
#ifdef _WIN32
			if (!vsync && (flags & ALLEGRO_OPENGL))
			{
				BOOL(APIENTRY *wglSwapInterval)(int) = (BOOL (APIENTRY *)(int))wglGetProcAddress("wglSwapIntervalEXT");
				if (wglSwapInterval)
				{
					wglSwapInterval(0);
				}
			}
#endif // _WIN32

			this->event_source = al_get_display_event_source(*this);
		}

		/// Returns the current width of the display
		point Width() const
		{
			return al_get_display_width();
		}

		/// Returns the current height of the display
		point Height() const
		{
			return al_get_display_height();
		}

		///Returns the display flags
		int Flags() const
		{
			int flags = al_get_new_display_flags();

			return flags;
		}

		/// Forces the thread to wait for VSYNC
		bool WaitForVSync()
		{
			this->Target();
			return al_wait_for_vsync();
		}

		/// Flips the display
		void Flip()
		{
			this->Target();
			al_flip_display();
		}

		/// Sets the global Allegro target display and bitmaps to this
		/// Shouldn't need to be called if using pure A5SES
		bool Target()
		{
			Bitmap::Target();
			return al_set_current_display(*this);
		}

		/// Sets the window title. No effect in fullscreen
		void SetTitle(const char *title)
		{
			this->Target();
			al_set_window_title(title);
		}

		/// Resizes the display
		bool Resize(point width, point height)
		{
			this->Target();
			return al_resize_display(width, height);
		}

		/// Acknowledges a Resize event
		bool AcknowledgeResize()
		{
			return al_acknowledge_resize(*this);
		}

		/// Destroys the display
		~Display()
		{
			al_destroy_display(*this);
		}

		virtual std::auto_ptr<a5::Event> Handle(ALLEGRO_EVENT *raw_event) const
		{
			std::auto_ptr<a5::Event> event(0);

			if (raw_event->any.source == this->event_source)
			{
				int subtype = a5::Event::Invalid;

				switch (raw_event->type)
				{
					case ALLEGRO_EVENT_DISPLAY_CLOSE:
						subtype = Event::Close;
						break;

					case ALLEGRO_EVENT_DISPLAY_RESIZE:
						subtype = Event::Resize;
						break;

					case ALLEGRO_EVENT_DISPLAY_EXPOSE:
						subtype = Event::Expose;
						break;

					case ALLEGRO_EVENT_DISPLAY_LOST:
						subtype = Event::Lost;
						break;

					case ALLEGRO_EVENT_DISPLAY_FOUND:
						subtype = Event::Found;
						break;

					case ALLEGRO_EVENT_DISPLAY_SWITCH_OUT:
						subtype = Event::SwitchOut;
						break;

					case ALLEGRO_EVENT_DISPLAY_SWITCH_IN:
						subtype = Event::SwitchIn;
						break;
				}

				event = std::auto_ptr<a5::Event>(new Event(this, subtype));
			}

			return event;
		}

		/// Returns a pointer to the Allegro display struture
		operator ALLEGRO_DISPLAY *()
		{
			return this->display;
		}
};

}

#endif // A5SES_DISPLAY_HPP_INCLUDED
