
/* A5SES is released under the zlib license.
 * See LICENSE.txt for more info.
 */

/// @file common.hpp
/// @ingroup Common
/// Contains classes and functions several a5ses components need

#ifndef A5SES_COMMON_HPP_INCLUDED
#define A5SES_COMMON_HPP_INCLUDED

#ifdef Rectangle
#undef Rectangle
#endif

#define Rectangle WINDOWS_Rectangle
#include <allegro5/allegro5.h>
#undef Rectangle

#ifdef A5SES_IMAGE_ADDON
#include <allegro5/allegro_image.h>
#endif

#ifdef RGB
#undef RGB
#ifndef WINDOWS_RGB
#define WINDOWS_RGB(r,g,b) ((COLORREF)(((BYTE)(r)|((WORD)((BYTE)(g))<<8))|(((DWORD)(BYTE)(b))<<16)))
#endif
#endif

/// @ingroup Common
/// A5SES version string
/// First 3 digits should match the Allegro version it's made for
#define A5SES_VERSION_STR "4.9.15.1"

/// @ingroup Common
/// A5SES major version
/// Should match the Allegro version it's made for
#define A5SES_MAJOR_VERSION 4

/// @ingroup Common
/// A5SES minor version
/// Should match the Allegro version it's made for
#define A5SES_MINOR_VERSION 9

/// @ingroup Common
/// A5SES WIP version
/// Should match the Allegro version it's made for
#define A5SES_WIP_VERSION 15

/// @ingroup Common
/// A5SES version
#define A5SES_VERSION 1

#if (A5SES_MAJOR_VERSION != ALLEGRO_VERSION) || (A5SES_MINOR_VERSION != ALLEGRO_SUB_VERSION) || (A5SES_WIP_VERSION != ALLEGRO_WIP_VERSION)
#warning Allegro version does not match the A5SES version
#endif

#include <exception>

/// Contains all A5SES classes and functions
namespace a5
{

/// @ingroup Common
/// Type used to represent bitmap/screen coordinates
#ifdef A5SES_INT_COORD
typedef int point;
#else
typedef float point;
#endif

/// @ingroup Common
/// Type used to represent angles
typedef float angle;

/// @ingroup Exception Common
/// Base class for A5SES exceptions
class Exception : public std::exception
{
	public:
		/// Returns the name of the exception as a string
		virtual const char *what()
		{
			return "a5::Exception";
		}
};

/// @ingroup Exception Common
/// Thrown when a function requiring a non-loaded addon is called
class Need_Addon : public Exception
{
	private:
		const char *addon_;

	public:
		/// Creates a Need_Addon exception
		Need_Addon(const char *addon__) : addon_(addon__) {}

		virtual const char *what()
		{
			return "a5::Need_Addon";
		}

		/// Returns the name of the missing addon
		const char *addon()
		{
			return this->addon_;
		}
};

/// @ingroup Exception Common
/// Thrown when an core Allegro module or addon fails to initialize
class Init_Failed : public Exception
{
	private:
		const char *addon_;

	public:
		/// Creates a Need_Addon exception
		Init_Failed(const char *addon__) : addon_(addon__) {}

		virtual const char *what()
		{
			return "a5::Init_Failed";
		}

		/// Returns the name of the missing addon
		const char *addon()
		{
			return this->addon_;
		}
};

/// @ingroup Common
/// Stores four coordinates representing a rectangle
class Rectangle
{
	public:
		/// X coordinate of top-left corner of rectangle
		point x1;

		/// Y coordinate of top-left corner of rectangle
		point y1;

		/// X coordinate of bottom-right corner of rectangle
		point x2;

		/// Y coordinate of bottom-right corner of rectangle
		point y2;

		/// Create a rectangle using the specified set of four coordinates
		Rectangle(point x1_, point y1_, point x2_, point y2_) : x1(x1_), y1(y1_), x2(x2_), y2(y2_) {}

		/// Returns the width of the rectangle
		point Width() const
		{
			return this->x2 - this->x1;
		}

		/// Returns the height of the rectangle
		point Height() const
		{
			return this->y2 - this->y1;
		}

		/// Returns the size of the rectangle (width * height)
		point Size() const
		{
			return this->Width() * this->Height();
		}
};

/// @internal
/// @ingroup Internal Common
/// Initializes the Allegro library
inline void _init()
{
	if (!al_get_system_driver())
	{
		if (!(al_init()))
		{
			Init_Failed("core");
		}

#ifdef A5SES_IMAGE_ADDON
		if (!(al_init_image_addon()))
		{
			throw Init_Failed("image");
		}
#endif
	}
}

}

#endif // A5SES_COMMON_HPP_INCLUDED
